from _Framework.CompoundComponent import CompoundComponent
from _Framework.SessionComponent import SessionComponent
from _Framework.ButtonElement import ButtonElement

import diagnostic
from LiveZoneScene import LiveZoneScene
from fingerboardcodes import FUNC_BLINKOFF, FUNC_BRIGHT
from colorlookup import sharedColorInfo, COLOR_GREEN

class LiveZone(SessionComponent):
	"""Represents the area of a Live session controlled from the Ztar fingerboard."""
	
	def __init__(self, nTracks, nScenes):
		# This is almost verbatim from the regular SessionComponent; we do it ourselves because if we
		# call the parent constructor the scenes will be populated with regular scenes, not our customized
		# LiveZoneScenes
		assert (SessionComponent._session_highlighting_callback != None)
		assert isinstance(nTracks, int)
		assert (nTracks >= 0)
		assert isinstance(nScenes, int)
		assert (nScenes >= 0)
		CompoundComponent.__init__(self)
		self._track_offset = -1
		self._scene_offset = -1
		self._num_tracks = nTracks
		self._bank_up_button = None
		self._bank_down_button = None
		self._bank_left_button = None
		self._bank_right_button = None
		self._stop_all_button = None
		self._next_scene_button = None
		self._prev_scene_button = None
		self._stop_track_clip_buttons = None
		self._scroll_up_ticks_delay = -1
		self._scroll_down_ticks_delay = -1
		self._scroll_right_ticks_delay = -1
		self._scroll_left_ticks_delay = -1
		self._stop_track_clip_value = 127
		self._offset_callbacks = []
		self._highlighting_callback = SessionComponent._session_highlighting_callback
		self._show_highlight = ((nTracks > 0) and (nScenes > 0))
		self._mixer = None
		self._selected_scene = self._create_scene(self._num_tracks)
		self.on_selected_scene_changed()
		self.register_components(self._selected_scene)
		self._scenes = []
		self._tracks_and_listeners = []
		for i in range(nScenes):
			self._scenes.append(self._create_scene(self._num_tracks))
			self.register_components(self._scenes[i])
		self.set_offsets(0,0)
		self._register_timer_callback(self._on_timer)
		return None
	
	def disconnect(self):
		SessionComponent.disconnect(self)
		return None
	
	def update(self):
		"""Override the refresh behavior of the session-controlled buttons. Since Ztar manages red-box scrolling
with special function codes, the turn_on() and turn_off() calls to those buttons in the default implementation 
don't apply."""
		if self._allow_updates:
			if self.is_enabled():
				sendValue = self._stop_track_clip_value
			else:
				sendValue = 0
			
			if self._stop_track_clip_buttons:
				for button in self._stop_track_clip_buttons:
					button.send_function(FUNC_BLINKOFF)
					button.send_function(FUNC_BRIGHT)
					button.send_value(sendValue)
			if self._stop_all_button:
				self._stop_all_button.send_function(FUNC_BLINKOFF)
				self._stop_all_button.send_function(FUNC_BRIGHT)
				self._stop_all_button.send_value(sendValue)
		else:
			self._update_requests += 1
	
	def syncWithLayout(self, layout, controls):
		diagnostic.logMessage("Resyncing: %d scenes, %d tracks to %d scenes, %d tracks" % (len(self._scenes), self._num_tracks, layout.nScenes, layout.nTracks))
		
		sceneDelta = layout.nScenes - len(self._scenes)
		trackDelta = layout.nTracks - self._num_tracks
		
		# Null out the old controls before we begin
		if self._stop_track_clip_buttons:
			for button in self._stop_track_clip_buttons:
				button.turn_off()
		self.set_stop_track_clip_buttons(None)
		self._stop_all_button.turn_off()
		self.set_stop_all_clips_button(None)
		
		# Resize and reposition the live zone using the new layout
		scenes = [LiveZoneScene(layout.nTracks, self.tracks_to_use) for iScene in range(layout.nScenes)]
		self.reinitWithScenes(scenes, layout.nTracks)
		
		if layout.scenesAreReversed():
			self.set_offsets(self._track_offset, self._scene_offset + (-1*sceneDelta))
		if layout.tracksAreReversed():
			self.set_offsets(self._track_offset + (-1*trackDelta), self._scene_offset)
		
		# Assign controls to the scenes, tracks, and clips that have a place in the new layout
		stopAllClipsLed = layout.stopAllClipsLed()
		stopClipsLeds = layout.stopClipsLeds()
		sceneLaunchLeds = layout.sceneLaunchLeds()
		clipLaunchLeds = [layout.clipLaunchLeds(i) for i in range(layout.nScenes)]
		
		diagnostic.logMessage('New stop-all-clips LED: %d' % stopAllClipsLed)
		diagnostic.logMessage('New stop-clips LEDs: %s' % stopClipsLeds)
		diagnostic.logMessage('New scene launch LEDs: %s' % sceneLaunchLeds)
		diagnostic.logMessage('New clip launch LEDs: %s' % clipLaunchLeds)
		
		newStopAllClips = controls[stopAllClipsLed]
		newStopClips = [controls[led] for led in stopClipsLeds]
		newSceneLaunches = [controls[led] for led in sceneLaunchLeds]
		newClipLaunches = [[controls[led] for led in clipLaunchLeds[i]] for i in range(layout.nScenes)]
		
		self.set_stop_all_clips_button(newStopAllClips)
		self.set_stop_track_clip_buttons(tuple(newStopClips))
		for i, scene in enumerate(self._scenes):
			scene.set_launch_button(newSceneLaunches[i])
			scene.setClipLaunchButtons(newClipLaunches[i])
	
	def reinitWithScenes(self, newScenes, nTracks):
		"""Reinitializes the session with the given scene components. As a side effect, the launch buttons associated
with the previous scene components and their clips are disconnected. This method does not attempt to remap them to
the new scenes."""
		self._num_tracks = nTracks
		
		for scene in self._scenes:
			self._sub_components.remove(scene)
			if scene._launch_button:
				scene._launch_button.turn_off()
			scene.set_launch_button(None)
			scene.setClipLaunchButtons([None] * len(scene._clip_slots))
		self._scenes = []
		
		if self._selected_scene and self._selected_scene in self._sub_components:
			self._sub_components.remove(self._selected_scene)
		self._selected_scene = self._create_scene(self._num_tracks)
		self.on_selected_scene_changed()
		self.register_components(self._selected_scene)
		
		for scene in newScenes:
			self._scenes.append(scene)
			self.register_components(scene)
			scene.set_default_value(sharedColorInfo['lookup'][COLOR_GREEN])
			scene.set_triggered_value(sharedColorInfo['lookup'][COLOR_GREEN])
			scene.update()
		
		self._reassign_tracks()
		self._reassign_scenes()
		self.set_show_highlight(False)
		self.set_show_highlight(True)
	
	def setSceneLaunchButtons(self, buttons):
		if len(buttons) < len(self._scenes):
			return
		
		for i, scene in enumerate(self._scenes):
			if scene._launch_button:
				scene._launch_button.turn_off()
			scene.set_launch_button(buttons[i])
	
	def _on_fired_slot_index_changed(self, index):
		"""Override to keep the stop track buttons lit. By default, the Live framework calls turn_off() on the stop
clips button for a track when a track for that clip is played."""
		if self.is_enabled() and self._stop_track_clip_buttons != None:
			self._stop_track_clip_buttons[index].send_value(self._stop_track_clip_value)
	
	def _create_scene(self, nTracks):
		return LiveZoneScene(nTracks, self.tracks_to_use)
	
	# Older versions of the Live Framework library use a property called _offset_callback instead
	# of the array, _offset_callbacks, used by Live 8.2. These next few methods define _offset_callback
	# as a property for compatibility with those previous versions.
	def _getOffsetCallback(self):
		if len(self._offset_callbacks) > 0:
			return self._offset_callbacks[0]
		else:
			return None
	
	def _setOffsetCallback(self, callback):
		if len(self._offset_callbacks) > 0:
			self._offset_callbacks[0] = callback
		else:
			self._offset_callbacks = [callback]
	
	def _delOffsetCallback(self):
		if len(self._offset_callbacks) > 0:
			del self._offset_callbacks[0]
	
	_offset_callback = property(_getOffsetCallback, _setOffsetCallback, _delOffsetCallback)
