import Live
from _Framework.ClipSlotComponent import ClipSlotComponent
from _Framework.ButtonElement import ButtonElement

import diagnostic
from FingerboardButton import FingerboardButton
from fingerboardcodes import *
from colorlookup import *

class LiveZoneClipSlot(ClipSlotComponent):
	"""Represents a clip slot controlled from the Live Zone on the Ztar fingerboard."""
	def __init__(self):
		ClipSlotComponent.__init__(self)
		
		self._lastColor = sharedColorInfo['lookup'][COLOR_LEDOFF]
		self.set_triggered_to_record_value(sharedColorInfo['lookup'][COLOR_RED])
		self.set_recording_value(sharedColorInfo['lookup'][COLOR_RED])
	
	def disconnect(self):
		if self._clip_slot != None:
			if self.has_clip():
				self._clip_slot.clip.remove_playing_status_listener(self._on_playing_state_changed)
				self._clip_slot.clip.remove_is_recording_listener(self._on_recording_state_changed)
				self._clip_slot.clip.remove_color_listener(self._on_color_changed)
			self._clip_slot.remove_has_clip_listener(self._on_clip_state_changed)
			self._clip_slot.remove_is_triggered_listener(self._on_slot_triggered_changed)
			self._clip_slot = None
		if self._launch_button:
			self._launch_button.remove_value_listener(self._launch_value)
			self._launch_button = None
	
	def update(self):
		"""Overrides the default implementation to include blink/brightness function codes in the button update."""
		self._has_fired_slot = False
		if self.is_enabled() and self._launch_button:
			valueToSend = None
			shouldBlink = False
			shouldBeBright = False
			if self._clip_slot:
				if self.has_clip():
					# If we have a clip, assume by default that we're stopped
					valueToSend = self._stopped_value
					shouldBlink = False
					shouldBeBright = False
					
					if self._clip_slot.clip.is_triggered:
						# We've been triggered, so blink and find out if we're recording or playing
						shouldBlink = True
						shouldBeBright = False
						if self._clip_slot.clip.will_record_on_start:
							valueToSend = self._triggered_to_record_value
						else:
							valueToSend = self._triggered_to_play_value
					elif self._clip_slot.clip.is_playing:
						# We're playing, so go full brightness and make sure we're not blinking
						shouldBlink = False
						shouldBeBright = True
						if self._clip_slot.clip.is_recording:
							valueToSend = self._recording_value
						else:
							valueToSend = self._started_value
				elif self._clip_slot.is_triggered:
					# We've been triggered, but we have no clip
					shouldBlink = True
					shouldBeBright = False
					if self._clip_slot.will_record_on_start:
						valueToSend = self._triggered_to_record_value
					else:
						valueToSend = self._triggered_to_play_value
				else:
					# No clip - so update with the off state
					shouldBlink = False
					shouldBeBright = False
					valueToSend = 0
				
				# Now (finally), update the button state
				if valueToSend:
					self._launch_button.send_value(valueToSend)
									
					if shouldBlink:
						self._launch_button.send_function(FUNC_BLINKON)
					else:
						self._launch_button.send_function(FUNC_BLINKOFF)
				
					if shouldBeBright:
						self._launch_button.send_function(FUNC_BRIGHT)
					else:
						self._launch_button.send_function(FUNC_DIM)
				else:
					self._launch_button.turn_off()
	
	def set_clip_slot(self, clip_slot):
		"""Overrides the default ClipSlotComponent method to add a color listener."""
		assert((clip_slot == None) or isinstance(clip_slot, Live.ClipSlot.ClipSlot))
		if clip_slot != self._clip_slot:
			if self._clip_slot != None:
				if self.has_clip():
					self._clip_slot.clip.remove_playing_status_listener(self._on_playing_state_changed)
					self._clip_slot.clip.remove_is_recording_listener(self._on_recording_state_changed)
					self._clip_slot.clip.remove_color_listener(self._on_color_changed)
				self._clip_slot.remove_is_triggered_listener(self._on_slot_triggered_changed)
				self._clip_slot.remove_playing_status_listener(self._on_playing_state_changed)
				self._clip_slot.remove_has_clip_listener(self._on_clip_state_changed)
				self._clip_slot.remove_controls_other_clips_listener(self._on_clip_state_changed)
			self._clip_slot = clip_slot
			if self._clip_slot != None:
				if self.has_clip():
					self._clip_slot.clip.add_playing_status_listener(self._on_playing_state_changed)
					self._clip_slot.clip.add_is_recording_listener(self._on_recording_state_changed)
					self._clip_slot.clip.add_color_listener(self._on_color_changed)
				self._clip_slot.add_is_triggered_listener(self._on_slot_triggered_changed)
				self._clip_slot.add_playing_status_listener(self._on_playing_state_changed)
				self._clip_slot.add_has_clip_listener(self._on_clip_state_changed)
				self._clip_slot.add_controls_other_clips_listener(self._on_clip_state_changed)
		self.setCurrentColor()
		self.update()
	
	def _on_color_changed(self):
		self.setCurrentColor()
		self.update()
	
	def setCurrentColor(self, colorcode=None):
		"""Sets the current color code of this clip slot to the specified color, or to the color of the current
clip if no color is specified."""
		if colorcode is None:
			if self._clip_slot and self.has_clip():
				colorcode = sharedColorInfo['lookup'].useColor(self._clip_slot.clip.color)
			else:
				colorcode = sharedColorInfo['lookup'][COLOR_WHITE]
		# Inform the lookup table that we're no longer using the last color, so it knows if it can free that address
		sharedColorInfo['lookup'].releaseColor(self._lastColor)
		self._lastColor = colorcode
		
		self.set_triggered_to_play_value(colorcode)
		self.set_started_value(colorcode)
		self.set_stopped_value(colorcode)
	
	def _on_clip_state_changed(self):
		"""Overrides the default ClipSlotComponent handler to change the color when the underlying clip changes."""
		assert(self._clip_slot != None)
		self.setCurrentColor()
		if self.has_clip():
			if not self._clip_slot.clip.playing_status_has_listener(self._on_playing_state_changed):
				self._clip_slot.clip.add_playing_status_listener(self._on_playing_state_changed)
			if not self._clip_slot.clip.is_recording_has_listener(self._on_recording_state_changed):
				self._clip_slot.clip.add_is_recording_listener(self._on_recording_state_changed)
		self.update()
		return None
	
	def _launch_value(self, value):
		"""Triggered when the launch button receives a value."""
		assert(self._launch_button != None)
		assert(value in range(128))
		if self.is_enabled() and self._clip_slot != None:
			objectToLaunch = self._clip_slot
			launchPressed = (value != 0) or not (self._launch_button.is_momentary())
			if self.has_clip():
				objectToLaunch = self._clip_slot.clip
			else:
				self._has_fired_slot = True
			
			if self.has_clip():
				diagnostic.logMessage('Clip slot %d: Setting fire state %d' % (self._launch_button._ledNumber, int(value != 0)))
				objectToLaunch.set_fire_button_state(value != 0)
				if launchPressed and self.song().select_on_launch:
					self.song().view.highlighted_clip_slot = self._clip_slot
			else:
				# If there's no clip in this slot, trigger the stop/record function
				self._clip_slot.fire()
		return None

