from _Framework.ControlSurface import ControlSurface
from _Framework.InputControlElement import *
from _Framework.ButtonElement import ButtonElement

import diagnostic
from LiveZone import LiveZone
from LiveZoneScene import LiveZoneScene
from LiveZoneClipSlot import LiveZoneClipSlot
from FingerboardButton import FingerboardButton
from fingerboardcodes import *
from livezonelayout import *
from colorlookup import *
import sysex

class Ztar(ControlSurface):
	def __init__(self, c_instance):
		ControlSurface.__init__(self, c_instance)
		diagnostic.logMessage = self.log_message
		
		self.set_suppress_rebuild_requests(True)
		self._session = None
		self._sessionLayout = LiveZoneLayout(NUM_STRINGS, NUM_FRETS)
		self._sessionLayout.onLayoutChange = self.turnOffUnusedLeds
		self._sentInitialColorSysex = False
		self.setupLiveZone()
		self.setupControlMap()
		self.set_suppress_rebuild_requests(False)
	
	def disconnect(self):
		self._session = None
		ControlSurface.disconnect(self)
		return None
	
	def refresh_state(self):
		self.log_message('Refreshing...')
		for control in self.controls:
			control.clear_send_cache()
		
		self._send_midi((144 + FUNCTION_CHANNEL, FUNCTIONCODES['clear-frame'], 0))
		# If the whole session needs to be refreshed, we can't assume the color lookup reflects the current
		# state of Live
		self.initializeColors()
		
		for component in self.components:
			component.update()
	
	def refreshLeds(self):
		self.log_message('Refreshing LEDs...')
		for led in self._sessionLayout.allActiveLeds():
			if led in self.ledControlMap:
				self.ledControlMap[led].resendLastValue()
	
	def setupControlMap(self):
		self.ledControlMap = {}
		isMomentary = True
		
		# Set up scene launch buttons
		sceneLaunchButtons = []
		for led in self._sessionLayout.sceneLaunchLeds():
			self.ledControlMap[led] = FingerboardButton(isMomentary, MIDI_NOTE_TYPE, led)
			sceneLaunchButtons.append(self.ledControlMap[led])
		
		# Set up stop track clips buttons
		stopTrackButtons = []
		for led in self._sessionLayout.stopClipsLeds():
			self.ledControlMap[led] = FingerboardButton(isMomentary, MIDI_NOTE_TYPE, led)
			stopTrackButtons.append(self.ledControlMap[led])
		self._session.set_stop_track_clip_value(sharedColorInfo['lookup'][COLOR_RED])
		self._session.set_stop_track_clip_buttons(tuple(stopTrackButtons))
		
		# Set up stop all clips button in the lower corner of the LiveZone
		stopAllLed = self._sessionLayout.stopAllClipsLed()
		self.ledControlMap[stopAllLed] = FingerboardButton(isMomentary, MIDI_NOTE_TYPE, stopAllLed)
		self._session.set_stop_all_clips_button(self.ledControlMap[stopAllLed])
		
		# Set up clip launch buttons
		for iScene in range(self._sessionLayout.nScenes):
			scene = self._session.scene(iScene)
			scene.set_launch_button(sceneLaunchButtons[iScene])
			scene.set_default_value(sharedColorInfo['lookup'][COLOR_GREEN])
			scene.set_triggered_value(sharedColorInfo['lookup'][COLOR_GREEN])
			
			clipLeds = self._sessionLayout.clipLaunchLeds(iScene)
			for iClipslot in range(self._sessionLayout.nTracks):
				clipslot = scene.clip_slot(iClipslot)
				
				led = clipLeds[iClipslot]
				self.ledControlMap[led] = FingerboardButton(isMomentary, MIDI_NOTE_TYPE, led)
				clipslot.set_launch_button(self.ledControlMap[led])
	
	def setupLiveZone(self):
		# Initialize the color lookup table - no callback at first, to give the clip slots a chance to register
		# their colors without the callback sending out a separate sysex for all of them
		sharedColorInfo['lookup'] = ColorLookup()
				
		# Create the session
		self.log_message("Creating session.")
		self.log_message("Rev 40.1")
		self.log_message("%d tracks, %d scenes" % (self._sessionLayout.nTracks, self._sessionLayout.nScenes))
		self._session = LiveZone(self._sessionLayout.nTracks, self._sessionLayout.nScenes)
		self._session.set_offsets(0, 0)
		
		# Button parameters
		isMomentary = True
		
		# Set up red box scroll buttons
		self._session.set_scene_bank_buttons(
			ButtonElement(isMomentary, MIDI_NOTE_TYPE, FUNCTION_CHANNEL, FUNCTIONCODES['scroll-down']),
			ButtonElement(isMomentary, MIDI_NOTE_TYPE, FUNCTION_CHANNEL, FUNCTIONCODES['scroll-up'])
		)
		self._session.set_track_bank_buttons(
			ButtonElement(isMomentary, MIDI_NOTE_TYPE, FUNCTION_CHANNEL, FUNCTIONCODES['scroll-right']),
			ButtonElement(isMomentary, MIDI_NOTE_TYPE, FUNCTION_CHANNEL, FUNCTIONCODES['scroll-left'])
		)
	
	def receive_midi(self, midiBytes):
		if len(midiBytes) >= 2:
			channel = midiBytes[0] % 16
			note = midiBytes[1]
			led = ledForChannelAndNote(channel, note)
			diagnostic.logMessage("Forwarding %s from LED %d to handler" % (str(midiBytes), led))
			
			# Force Note-Off velocity to 0
			if len(midiBytes) == 3 and midiBytes[0] & 0xf0 == 128:
				midiBytes = (midiBytes[0], midiBytes[1], 0)
			ControlSurface.receive_midi(self, midiBytes)
	
	def syncLiveZone(self):
		"""Reconfigures the session object and its controls to match the layout specified in _sessionLayout."""
		# Adjust red box and launch buttons
		self._session.syncWithLayout(self._sessionLayout, self.ledControlMap)
	
	def turnOffUnusedLeds(self, sceneLaunchTuple, stopClipsTuple, stopAllTuple, clipLaunchTuple):
		pass
		# for led in self._sessionLayout.allLeds():
			# diagnostic.logMessage('Turning out %d' % led)
			# channel, note = channelAndNoteForLed(led)
			# self._send_midi(tuple((144+channel, note, 0)))
			# self.ledControlMap[led].turn_off()

	def initializeColors(self):
		self._sentInitialColorSysex = False
		self.sendColorSysex([])		# No need to specify colors here, since the first call will send the whole table
		sharedColorInfo['lookup'].addColorCallback(self.sendColorSysex)
	
	def sendColorSysex(self, addedColors):
		"""Responds when a color is added to the shared lookup table by one of the clip slots."""
		if not self._sentInitialColorSysex:
			# If we haven't sent a color sysex yet, then whatever's in the lookup table needs to be sent. Pack it
			# all into one message so as not to annoy the Ztar with a hundred tiny messages.
			
			# Allocate 3 bytes for each mapped color
			rgbColors, lookups = zip(*[[rgbColor, lookup] for rgbColor, lookup in sharedColorInfo['lookup']])
			colorPackets = [0] * (max(lookups) * 3)
			for i, color in enumerate(rgbColors):
				red, green, blue = separateRgb(color)
				red, green, blue = downscaleColorValue(red, green, blue)	# Downscale from 8 bits to 5
				offset = lookups[i]*3
				colorPackets[offset:offset+3] = [red, green, blue]
			
			message = sysex.sysexFromPackets(
				[0x00, 0x01, 0x42],		# Starr Labs manufacturer ID
				[0x40, 0x00, 0x00],		# Packet ID - 0x40 for color message, 0x0000 for LED-FB type 0
				colorPackets
			)
			
			self._send_midi(tuple(message))
			self._sentInitialColorSysex = True
		else:
			# Allocate 3 bytes for each color in the update
			colorPackets = []
			for color in addedColors:
				red, green, blue = separateRgb(color)
				red, green, blue = downscaleColorValue(red, green, blue)	# Downscale from 8 bits to 5
				lookup = sharedColorInfo['lookup'][color]
				colorPackets.extend([lookup, red, green, blue])
			
			message = sysex.sysexFromPackets(
				[0x00, 0x01, 0x42],		# Starr Labs manufacturer ID
				[0x40, 0x00, 0x01],		# Packet ID - 0x40 for color message, 0x0000 for LED-FB type 1
				colorPackets
			)
			
			self._send_midi(tuple(message))
	
	def handle_sysex(self, midi_bytes):
		# The only message handled here, at this point, is the message to resize/reorient the Live Zone.
		self.log_message("Handling sysex..." + str(midi_bytes))
		newLiveZone = sysex.handleLiveZoneSysex(midi_bytes)
		if not newLiveZone:
			if sysex.isStarrLabsSysex(midi_bytes):
				self.refreshLeds()
			return
		
		# Coerce out-of-range values to the minimum and maximum the hardware allows
		if newLiveZone['startFret'] < 0:
			newLiveZone['startFret'] = 0
		if newLiveZone['startString'] < 0:
			newLiveZone['startString'] = 0
		if newLiveZone['endFret'] > NUM_FRETS - 1:
			newLiveZone['endFret'] = NUM_FRETS - 1
		if newLiveZone['endString'] > NUM_STRINGS - 1:
			newLiveZone['endString'] = NUM_STRINGS - 1
		
		if newLiveZone and newLiveZone['endFret'] - newLiveZone['startFret'] > 0 and newLiveZone['endString'] - newLiveZone['startString'] > 0:
			self.log_message("Remapping live zone")
			if newLiveZone['isRightHanded']:
				self._sessionLayout.orientLiveZone(HIGH_FRET, LOW_STRING)
			else:
				self._sessionLayout.orientLiveZone(LOW_FRET, HIGH_STRING)
			
			if self._sessionLayout.scenesAreFrets():
				self._sessionLayout.repositionLiveZone(newLiveZone['startFret'], newLiveZone['endFret'],
														newLiveZone['startString'], newLiveZone['endString'])
			else:
				self._sessionLayout.repositionLiveZone(newLiveZone['startString'], newLiveZone['endString'],
														newLiveZone['startFret'], newLiveZone['endFret'])
			
			self.syncLiveZone()
			self.refreshLeds()
