NUM_STRINGS = 6
NUM_FRETS = 24

BASE_CHANNEL = 14	# Actually MIDI channel 15; code indexes channels starting at 0
STRINGS_PER_ZONE = 5
LEDS_PER_CHANNEL = STRINGS_PER_ZONE * NUM_FRETS

def channelForLed(led):
	return int(led / LEDS_PER_CHANNEL) + BASE_CHANNEL

def noteForLed(led):
	return led % LEDS_PER_CHANNEL

def channelAndNoteForLed(led):
	return (channelForLed(led), noteForLed(led))

def ledForChannelAndNote(channel, note):
	return (channel - BASE_CHANNEL) * LEDS_PER_CHANNEL + note

# Constants used to designate the sides of the LiveZone. "Low" and "high" are used with reference to the LED numbers;
# LEDs on the "low" fret have lower numbers than those on the "high" fret. 
# This may or may not correspond to the "low" or "high" musical notes on the instrument itself. 
LOW_STRING = 0
HIGH_STRING = 1
LOW_FRET = 2
HIGH_FRET = 3

class LiveZoneLayout(object):
	"""Class for representing the layout of a LiveZone, a collection of LEDs used to launch clips, scenes, and tracks
in Ableton Live."""
	def __init__(self, nStrings=NUM_STRINGS, nFrets=NUM_FRETS):
		# Declare an optional callback for layout changes
		self.onLayoutChange = None
		
		self.ledCount = nStrings * nFrets
		self.nStrings = nStrings
		self.nFrets = nFrets
		
		# Default right-handed arrangement
		self.sceneOffset = 0
		self.trackOffset = 0
		self.resizeLiveZone(self.nStrings - 1, self.nFrets - 1)
		self.orientLiveZone(sceneLaunch=HIGH_FRET, stopClips=LOW_STRING)
	
	def orientLiveZone(self, sceneLaunch, stopClips):
		"""Marks locations of specific LEDs in the LiveZone, so this instance knows which LEDs correspond to which
function, scene, or track. We're assuming that the clip in the first row and first column of the Ableton display will
be represented by the LED in the corner opposite the intersection of the scene-launch and stop-clips arrays."""
		if self.onLayoutChange:
			oldSceneLaunches = self.sceneLaunchLeds()
			oldStopClips = self.stopClipsLeds()
			oldStopAll = self.stopAllClipsLed()
			oldClipLaunches = [self.clipLaunchLeds(i) for i in range(self.nScenes)]
		
		self._sceneLaunchPlacement = sceneLaunch
		self._stopClipsPlacement = stopClips
		if self.scenesAreFrets():
			self.nScenes = min(self.nScenes, self.nFrets)
			self.nTracks = min(self.nTracks, self.nStrings)
		else:
			self.nScenes = min(self.nScenes, self.nStrings)
			self.nTracks = min(self.nTracks, self.nFrets)
		
		if self.onLayoutChange:
			newSceneLaunches = self.sceneLaunchLeds()
			newStopClips = self.stopClipsLeds()
			newStopAll = self.stopAllClipsLed()
			newClipLaunches = [self.clipLaunchLeds(i) for i in range(self.nScenes)]
			self.onLayoutChange(
				(oldSceneLaunches, newSceneLaunches),
				(oldStopClips, newStopClips),
				(oldStopAll, newStopAll),
				(oldClipLaunches, newClipLaunches)
			)
	
	def resizeLiveZone(self, nScenes, nTracks):
		if self.onLayoutChange:
			oldSceneLaunches = self.sceneLaunchLeds()
			oldStopClips = self.stopClipsLeds()
			oldStopAll = self.stopAllClipsLed()
			oldClipLaunches = [self.clipLaunchLeds(i) for i in range(self.nScenes)]
		
		self.nScenes = nScenes
		self.nTracks = nTracks
		
		if self.onLayoutChange:
			newSceneLaunches = self.sceneLaunchLeds()
			newStopClips = self.stopClipsLeds()
			newStopAll = self.stopAllClipsLed()
			newClipLaunches = [self.clipLaunchLeds(i) for i in range(self.nScenes)]
			self.onLayoutChange(
				(oldSceneLaunches, newSceneLaunches),
				(oldStopClips, newStopClips),
				(oldStopAll, newStopAll),
				(oldClipLaunches, newClipLaunches)
			)
	
	def repositionLiveZone(self, startScene, endScene, startTrack, endTrack):
		if self.onLayoutChange:
			oldSceneLaunches = self.sceneLaunchLeds()
			oldStopClips = self.stopClipsLeds()
			oldStopAll = self.stopAllClipsLed()
			oldClipLaunches = [self.clipLaunchLeds(i) for i in range(self.nScenes)]
		
		self.sceneOffset = startScene
		self.trackOffset = startTrack
		self.nScenes = endScene - startScene
		self.nTracks = endTrack - startTrack
		
		if self.onLayoutChange:
			newSceneLaunches = self.sceneLaunchLeds()
			newStopClips = self.stopClipsLeds()
			newStopAll = self.stopAllClipsLed()
			newClipLaunches = [self.clipLaunchLeds(i) for i in range(self.nScenes)]
			self.onLayoutChange(
				(oldSceneLaunches, newSceneLaunches),
				(oldStopClips, newStopClips),
				(oldStopAll, newStopAll),
				(oldClipLaunches, newClipLaunches)
			)
	
	def allLeds(self):
		"""Returns a list of all the LED numbers this instance could ever cover."""
		return range(0, NUM_STRINGS * NUM_FRETS)
	
	def allActiveLeds(self):
		"""Returns a list of all the LED numbers attached to a function in the current configuration."""
		sceneLaunches = self.sceneLaunchLeds()
		stopClips = self.stopClipsLeds()
		stopAll = [self.stopAllClipsLed()]
		clipLaunches = [self.clipLaunchLeds(i) for i in range(self.nScenes)]
		leds = sceneLaunches + stopClips + stopAll
		for clipArray in clipLaunches:
			leds.extend(clipArray)
		return leds

	def ledsOfString(self, string):
		offset = string * self.nFrets
		return [led for led in range(offset, offset + self.nFrets)]
	
	def ledsOfFret(self, fret):
		return [led for led in range(fret, fret + (self.nStrings * self.nFrets), self.nFrets)]
	
	def ledsAlongSide(self, stringOrFret):
		"""Returns a list of LED numbers along the specified string or fret."""
		if stringOrFret == LOW_STRING:
			if self.scenesAreFrets():
				return self.ledsOfString(self.trackOffset)
			else:
				return self.ledsOfString(self.sceneOffset)
		elif stringOrFret == HIGH_STRING:
			if self.scenesAreFrets():
				return self.ledsOfString(self.trackOffset + self.nTracks)
			else:
				return self.ledsOfString(self.sceneOffset + self.nScenes)
		elif stringOrFret == LOW_FRET:
			if self.scenesAreFrets():
				return self.ledsOfFret(self.sceneOffset)
			else:
				return self.ledsOfFret(self.trackOffset)
		elif stringOrFret == HIGH_FRET:
			if self.scenesAreFrets():
				return self.ledsOfFret(self.sceneOffset + self.nScenes)
			else:
				return self.ledsOfFret(self.trackOffset + self.nTracks)
	
	def scenesAreFrets(self):
		return self._sceneLaunchPlacement == LOW_STRING or self._sceneLaunchPlacement == HIGH_STRING
	
	def tracksAreFrets(self):
		return self._stopClipsPlacement == LOW_STRING or self._stopClipsPlacement == HIGH_STRING
	
	def scenesAreReversed(self):
		return self._stopClipsPlacement == LOW_STRING or self._stopClipsPlacement == LOW_FRET
	
	def tracksAreReversed(self):
		return self._sceneLaunchPlacement == LOW_STRING or self._sceneLaunchPlacement == LOW_FRET
	
	def stopAllClipsLed(self):
		sceneLaunchSet = set(self.ledsAlongSide(self._sceneLaunchPlacement))
		stopClipsSet = set(self.ledsAlongSide(self._stopClipsPlacement))
		led = sceneLaunchSet.intersection(stopClipsSet).pop()
		return led
	
	def sceneLaunchLeds(self):
		lastLed = self.sceneOffset + self.nScenes + 1
		leds = self.ledsAlongSide(self._sceneLaunchPlacement)[self.sceneOffset:lastLed]
		if self._stopClipsPlacement == LOW_STRING or self._stopClipsPlacement == LOW_FRET:
			leds = leds[1:]
		elif self._stopClipsPlacement == HIGH_STRING or self._stopClipsPlacement == HIGH_FRET:
			leds = leds[:-1]
		
		if self.scenesAreReversed():
			leds.reverse()
		return leds
	
	def stopClipsLeds(self):
		lastLed = self.trackOffset + self.nTracks + 1
		leds = self.ledsAlongSide(self._stopClipsPlacement)[self.trackOffset:lastLed]
		if self._sceneLaunchPlacement == LOW_STRING or self._sceneLaunchPlacement == LOW_FRET:
			leds = leds[1:]
		elif self._sceneLaunchPlacement == HIGH_STRING or self._sceneLaunchPlacement == HIGH_FRET:
			leds = leds[:-1]
		
		if self.tracksAreReversed():
			leds.reverse()
		return leds
	
	def clipLaunchLeds(self, iScene):
		if self.scenesAreFrets():
			if self.scenesAreReversed():
				leds = self.ledsOfFret(self.sceneOffset + self.nTracks - iScene)
			else:
				leds = self.ledsOfFret(self.sceneOffset + iScene)
		else:
			if self.scenesAreReversed():
				leds = self.ledsOfString(self.sceneOffset + self.nScenes - iScene)
			else:
				leds = self.ledsOfString(self.sceneOffset + iScene)
		
		if self.tracksAreReversed():
			leds.reverse()
			return leds[-1 - self.trackOffset - self.nTracks:-1 - self.trackOffset]
		else:
			lastLed = self.trackOffset + self.nTracks
			return leds[self.trackOffset:lastLed]
