MFRID_STARRLABS = (0x00, 0x01, 0x42)
PACKETTYPE_LIVE = 0x40
PACKET_REMAP_COLORS = 0
PACKET_CHANGE_COLOR = 1
PACKET_FRAMEBUFFER = 2
PACKET_LIVEZONE = 3

def make14BitInt(anInt, msb=True):
	"""Makes a 14-bit integer (2 bytes, high bit unused) from an integer. Used in forming our sysex packets."""
	anInt = anInt % 16384
	lowByte = anInt % 128
	highByte = (anInt - lowByte) >> 7
	if msb:
		return [highByte, lowByte]
	else:
		return [lowByte, highByte]

def valueOf14BitInt(bytes):
	"""Returns an integer represented by joining the two given bytes, leaving off the high bit of each."""
	maskedBytes = [byte & 0x7f for byte in bytes]
	theInt = maskedBytes[0] << 7
	theInt += maskedBytes[1]
	return theInt

def checksumWithMod(dataBytes, modulus):
	"""Computes a checksum with the specified modulus (2^number of checksum bits)."""
	return sum(dataBytes) % modulus

def sysexFromPackets(mfrId, msgId, packetBytes):
	"""Creates a sysex message with the given manufacturer and message ID bytes, computing a length and checksum,
and appending the sysex header and trailer bytes."""
	msgIdOffset = 1 + len(mfrId)
	msgLenOffset = msgIdOffset + len(msgId)
	packetsOffset = msgLenOffset + 2
	checksumOffset = packetsOffset + len(packetBytes)
	msgLength = checksumOffset + 2
	sysex = [0] * msgLength
	
	sysex[0] = 0xf0		# SysEx header
	sysex[1:msgIdOffset] = mfrId
	sysex[msgIdOffset:msgLenOffset] = msgId
	sysex[msgLenOffset:packetsOffset] = make14BitInt(len(packetBytes))
	sysex[packetsOffset:checksumOffset] = packetBytes
	sysex[checksumOffset] = checksumWithMod(sysex, 128)
	sysex[-1] = 0xf7	# SysEx trailer
	
	return sysex

def isStarrLabsSysex(sysex):
	if sysex[0] != 0xf0:
		return False
	elif sysex[1:4] != MFRID_STARRLABS:
		return False
	elif sysex[4] != PACKETTYPE_LIVE:
		return False
	elif sysex[-1] != 0xf7:
		return False
	else:
		return True

def isStarrLabsPacketType(sysex, packetType, packetSubtype):
	"""Returns True if the given byte list is formatted like a LiveZone SysEx message."""
	if sysex[0] != 0xf0:
		return False
	elif sysex[1:4] != MFRID_STARRLABS:
		return False
	elif sysex[4] != packetType:
		return False
	elif valueOf14BitInt(sysex[5:7]) != packetSubtype:
		return False
	elif valueOf14BitInt(sysex[7:9]) != len(sysex[9:-2]):
		return False
	elif sysex[-2] != checksumWithMod(sysex[1:-2], 128):
		return False
	elif sysex[-1] != 0xf7:
		return False
	else:
		return True

def extractPayload(sysex):
	return sysex[9:-2]

def handleLiveZoneSysex(sysex):
	"""Attempts to parse the given list of bytes as a LiveZone SysEx message, returning a dictionary containing
the message's information, or None if the message is invalid."""
	if not isStarrLabsPacketType(sysex, PACKETTYPE_LIVE, PACKET_LIVEZONE):
		return None
	
	payload = extractPayload(sysex)
	nStrings = payload[3] - payload[1]
	nFrets = payload[4] - payload[2]
	isRightHanded = payload[5] == 0
	return {
		'startString': payload[1],
		'endString': payload[3],
		'startFret': payload[2],
		'endFret': payload[4],
		'isRightHanded': isRightHanded
	}
